/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <sys/mman.h>
#include <dlog.h>
#include <Elementary.h>
#include <Ecore_Wayland.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <tbm_surface.h>
#include <tbm_surface_internal.h>
#include <uuid/uuid.h>

#include "screen_connector_toolkit.h"
#include "screen_connector_toolkit_evas.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER_EVAS"
#define SC_TOOLKIT_HANDLE_TAG "SC_TOOLKIT_HANDLE_TAG"

struct _screen_connector_type_evas_h {
	GHashTable *toolkit_table;
	Evas_Object *viewer_win;
};

struct _cur_buffer_info_h {
	struct wl_buffer *cur_buffer;
	int ref_count;
};
typedef struct _cur_buffer_info_h *cur_buffer_info_h;

struct _viewer_visibility_h {
	visibility_type visibility;
	int win_id;
};
typedef struct _viewer_visibility_h *viewer_visibility_h;

struct _screen_connector_toolkit_evas_h {
	screen_connector_toolkit_evas_ops *ops;
	screen_connector_toolkit_h toolkit_h;
	int freeze;
	bool is_init;
	bool is_down;
	Evas_Object *img_tbm;
	Evas_Object *img_file;
	uint32_t img_type;
	struct wl_buffer *cur_buffer;
	struct _screen_connector_type_evas_h *type_h;
	guint resuming_timer;
	GList *viewer_visibility_list;
	visibility_type auto_visibility;
	int bind_win_id;
	screen_connector_screen_type_e screen_type;
	bool flick;
	int prev_y;
	void *data;
	bool blocked;
};

static GHashTable *__cur_buffer_table = NULL;
static GHashTable *__type_table = NULL;
static Ecore_Event_Handler *__visibility_listener;
static int __delayed_resuming_time;

static void __rs_cb_mouse_down(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_mouse_up(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_touch_multi_down(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_touch_multi_up(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_mouse_move(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_mouse_wheel(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_show(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_hide(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __rs_cb_resize(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void __obj_update_visibility(void *data, Evas *e, Evas_Object *obj, void *event_info);

static void __cancel_touch(screen_connector_toolkit_evas_h h)
{
	struct tizen_remote_surface *surface;

	if (!h->is_down)
		return;

	h->is_down = false;
	surface = screen_connector_toolkit_get_trs(h->toolkit_h);
	if (!surface) {
		LOGE("failed to get tizen remote surface");
		return;
	}

	tizen_remote_surface_transfer_touch_cancel(surface);
}

static gint __comp_win_id(gconstpointer a, gconstpointer b)
{
	viewer_visibility_h win = (viewer_visibility_h)a;

	return !(win->win_id == GPOINTER_TO_INT(b));
}

static int __set_window_visibility_info(
		screen_connector_toolkit_evas_h h,
		int win_id, visibility_type visibility)
{
	viewer_visibility_h visibility_h;
	GList *tmp_list = NULL;

	if (h->viewer_visibility_list)
		tmp_list = g_list_find_custom(
					h->viewer_visibility_list,
					GINT_TO_POINTER(win_id), __comp_win_id);

	if (tmp_list == NULL) {
		visibility_h = (viewer_visibility_h)calloc(1,
				sizeof(struct _viewer_visibility_h));
		if (visibility_h == NULL) {
			LOGE("Out of memory fail to alloc");
			return -1;
		}
		h->viewer_visibility_list =
				g_list_append(
					h->viewer_visibility_list,
					visibility_h);
	} else {
		visibility_h = (viewer_visibility_h)tmp_list->data;
	}
	visibility_h->win_id = win_id;
	visibility_h->visibility = visibility;

	LOGI("set win(%d) to (%d), len(%d)", win_id, visibility, g_list_length(h->viewer_visibility_list));

	return 0;
}

static int __remove_window_visibility_info(
		screen_connector_toolkit_evas_h h,
		int win_id)
{
	viewer_visibility_h visibility_h;
	GList *tmp_list;

	tmp_list = g_list_find_custom(
				h->viewer_visibility_list,
				GINT_TO_POINTER(win_id), __comp_win_id);

	if (tmp_list == NULL) {
		LOGE("win id %d not exist");
		return -1;
	}

	visibility_h = (viewer_visibility_h)tmp_list->data;
	h->viewer_visibility_list = g_list_remove(
						h->viewer_visibility_list,
						visibility_h);
	free(visibility_h);

	LOGI("remove win(%d) len(%d)", win_id, g_list_length(h->viewer_visibility_list));

	return 0;
}

static void __cur_buffer_info_ref(cur_buffer_info_h info)
{
	info->ref_count++;
}

static bool __cur_buffer_info_unref(struct tizen_remote_surface *trs, cur_buffer_info_h info)
{
	tbm_surface_h cur_tbm_surface;

	info->ref_count--;
	if (info->ref_count == 0) {
		tizen_remote_surface_release(trs, info->cur_buffer);
		cur_tbm_surface = (tbm_surface_h)wl_buffer_get_user_data(info->cur_buffer);
		tbm_surface_internal_unref(cur_tbm_surface);
		g_hash_table_remove(__cur_buffer_table, info->cur_buffer);
		return true;
	}

	return false;
}

static void __destroy_cur_buffer(gpointer data)
{
	cur_buffer_info_h info = (cur_buffer_info_h)data;

	if (!info)
		return;

	if (info->cur_buffer)
		wl_buffer_destroy(info->cur_buffer);

	free(info);
}

static void __clear_img_tbm(screen_connector_toolkit_evas_h toolkit_evas_h)
{
	char *plug_id;

	if (toolkit_evas_h->img_tbm == NULL)
		return;

	plug_id = evas_object_data_del(toolkit_evas_h->img_tbm, "___PLUGID");
	if (plug_id)
		free(plug_id);

	evas_object_del(toolkit_evas_h->img_tbm);
	toolkit_evas_h->img_tbm = NULL;
}

static void __clear_img_file(screen_connector_toolkit_evas_h toolkit_evas_h)
{
	if (toolkit_evas_h->img_file == NULL)
		return;

	evas_object_del(toolkit_evas_h->img_file);
	toolkit_evas_h->img_file = NULL;
}

static void __destroy_toolkit_evas_h(gpointer data)
{
	char *plug_id;
	cur_buffer_info_h c_buf_info;
	screen_connector_toolkit_evas_h toolkit_evas_h = data;
	struct tizen_remote_surface *surface;

	if (!toolkit_evas_h)
		return;

	if (toolkit_evas_h->img_tbm != NULL) {

		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_DOWN,
				__rs_cb_mouse_down);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_UP,
				__rs_cb_mouse_up);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MULTI_DOWN,
				__rs_cb_touch_multi_down);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MULTI_UP,
				__rs_cb_touch_multi_up);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_MOVE,
				__rs_cb_mouse_move);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_WHEEL,
				__rs_cb_mouse_wheel);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_SHOW,
				__rs_cb_show);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_HIDE,
				__rs_cb_hide);
		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_RESIZE,
				__rs_cb_resize);

		evas_object_event_callback_del(
				toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOVE,
				__obj_update_visibility);

		c_buf_info = g_hash_table_lookup(__cur_buffer_table,
				toolkit_evas_h->cur_buffer);
		if (c_buf_info != NULL) {
			surface = screen_connector_toolkit_get_trs(
						toolkit_evas_h->toolkit_h);
			__cur_buffer_info_unref(surface, c_buf_info);
		}

		plug_id = evas_object_data_del(toolkit_evas_h->img_tbm, "___PLUGID");
		if (plug_id)
			free(plug_id);

		__clear_img_tbm(toolkit_evas_h);
	}

	if (toolkit_evas_h->toolkit_h)
		screen_connector_toolkit_remove(toolkit_evas_h->toolkit_h);

	if (toolkit_evas_h->ops)
		free(toolkit_evas_h->ops);

	if (toolkit_evas_h->resuming_timer > 0)
		g_source_remove(toolkit_evas_h->resuming_timer);

	if (toolkit_evas_h->viewer_visibility_list)
		g_list_free_full(toolkit_evas_h->viewer_visibility_list, free);

	free(toolkit_evas_h);
}

static void __destroy_type_h(gpointer data)
{
	GHashTableIter iter;
	gpointer key, value;
	screen_connector_type_evas_h type_h = (screen_connector_type_evas_h)data;
	if (!type_h)
		return;

	if (type_h->toolkit_table) {
		g_hash_table_iter_init(&iter, type_h->toolkit_table);
		while (g_hash_table_iter_next(&iter, &key, &value))
			__destroy_toolkit_evas_h(value);

		g_hash_table_destroy(type_h->toolkit_table);
		type_h->toolkit_table = NULL;
	}
	free(type_h);
}

EXPORT_API int screen_connector_toolkit_evas_init(Evas_Object *win, screen_connector_screen_type_e type)
{
	screen_connector_type_evas_h type_h;

	if (__type_table == NULL)
		__type_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, __destroy_type_h);

	if (__cur_buffer_table == NULL)
		__cur_buffer_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, __destroy_cur_buffer);

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {

		type_h = (screen_connector_type_evas_h)calloc(1, sizeof(struct _screen_connector_type_evas_h));
		g_hash_table_insert(__type_table, GINT_TO_POINTER(type), type_h);

		type_h->toolkit_table = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
		if (!type_h->toolkit_table) {
			LOGE("failed to create table");
			return -1;
		}
	}

	type_h->viewer_win = win;
	return screen_connector_toolkit_init(type);
}

EXPORT_API int screen_connector_toolkit_evas_fini(screen_connector_screen_type_e type)
{
	screen_connector_type_evas_h type_h;
	GHashTableIter iter;
	gpointer key, value;

	if (__type_table == NULL) {
		LOGI("__type_table is NULL");
		return 0;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h) {
		if (type_h->toolkit_table) {
			g_hash_table_iter_init(&iter, type_h->toolkit_table);
			while (g_hash_table_iter_next(&iter, &key, &value))
				__destroy_toolkit_evas_h(value);
			g_hash_table_destroy(type_h->toolkit_table);
			type_h->toolkit_table = NULL;
		}
		g_hash_table_remove(__type_table, GINT_TO_POINTER(type));
	}

	if (g_hash_table_size(__type_table) == 0) {
		g_hash_table_destroy(__type_table);
		__type_table = NULL;

		if (__cur_buffer_table)
			g_hash_table_destroy(__cur_buffer_table);
		__cur_buffer_table = NULL;
	}

	__delayed_resuming_time = 0;

	return screen_connector_toolkit_fini(type);
}

static bool __obj_is_visible(screen_connector_toolkit_evas_h toolkit_evas_h)
{
	int x, y, w, h;
	Ecore_Wl_Window *window = NULL;
	int window_x, window_y, window_w, window_h;
	int rotation;

	if (toolkit_evas_h->type_h == NULL || toolkit_evas_h->type_h->viewer_win == NULL) {
		LOGE("type_h is NULL");
		return false;
	}

	if (toolkit_evas_h->img_tbm == NULL) {
		LOGE("img_tbm is NULL");
		return false;
	}

	window = elm_win_wl_window_get(toolkit_evas_h->type_h->viewer_win);
	ecore_wl_window_geometry_get(window, &window_x, &window_y, &window_w, &window_h);
	evas_object_geometry_get(toolkit_evas_h->img_tbm, &x, &y, &w, &h);
	rotation = ecore_wl_window_rotation_get(window);

	if (x >= 0 && x < window_w &&
			y >= 0 && y < window_h &&
			(rotation == 0 || rotation == 180)) {
		LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
				x, y, w, h, window_w, window_h, rotation);
		return true;
	} else if (x >= 0 && x < window_h &&
			y >= 0 && y < window_w &&
			(rotation == 90 || rotation == 270)) {
		LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
				x, y, w, h, window_w, window_h, rotation);
		return true;
	}

	return false;
}

static gboolean __resuming_timeout_cb(gpointer user_data)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = user_data;
	struct tizen_remote_surface *surface;

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface) {
		LOGI("delayed resume");
		tizen_remote_surface_transfer_visibility(surface,
				TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_VISIBLE);
	}

	toolkit_evas_h->resuming_timer = 0;

	return G_SOURCE_REMOVE;
}

static int __set_visibility(screen_connector_toolkit_evas_h toolkit_evas_h, visibility_type type)
{
	int obscured;
	struct tizen_remote_surface *surface;

	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}

	switch (type) {
	case VISIBILITY_TYPE_UNOBSCURED:
	case VISIBILITY_TYPE_PARTIALLY_OBSCURED:
		obscured = TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_VISIBLE;
		break;
	case VISIBILITY_TYPE_FULLY_OBSCURED:
		obscured = TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_INVISIBLE;
		break;
	default:
		return -1;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (!surface)
		return -1;

	if (__delayed_resuming_time == 0 ||
		toolkit_evas_h->screen_type == SCREEN_CONNECTOR_SCREEEN_TYPE_WATCH) {
		if (obscured == TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_INVISIBLE)
			__cancel_touch(toolkit_evas_h);

		LOGI("normal visibility %d", obscured);
		tizen_remote_surface_transfer_visibility(surface, obscured);

		return 0;
	}

	if (obscured == TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_INVISIBLE) {
		if (toolkit_evas_h->resuming_timer > 0) {
			g_source_remove(toolkit_evas_h->resuming_timer);
			toolkit_evas_h->resuming_timer = 0;
		}

		__cancel_touch(toolkit_evas_h);
		tizen_remote_surface_transfer_visibility(surface, obscured);
	} else {
		if (toolkit_evas_h->resuming_timer == 0) {
			toolkit_evas_h->resuming_timer = g_timeout_add(__delayed_resuming_time,
					__resuming_timeout_cb, toolkit_evas_h);
		}
	}

	return 0;
}

static void __obj_update_visibility(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;

	if (!toolkit_evas_h->is_init) {
		toolkit_evas_h->is_init = true;
		return;
	}

	if (__obj_is_visible(toolkit_evas_h))
		toolkit_evas_h->auto_visibility = VISIBILITY_TYPE_UNOBSCURED;
	else
		toolkit_evas_h->auto_visibility = VISIBILITY_TYPE_FULLY_OBSCURED;

	__set_visibility(toolkit_evas_h, toolkit_evas_h->auto_visibility);
}

static void __rs_cb_mouse_down(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Down *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("mouse down: %d %d", ev->canvas.x - x, ev->canvas.y - y);
	toolkit_evas_h->flick = false;
	toolkit_evas_h->prev_y = ev->canvas.y - y;

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_UP_DOWN)) {
		LOGD("mouse down disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	toolkit_evas_h->is_down = true;
	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_DOWN,
			0,
			ev->button,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_mouse_up(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Up *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("mouse up: %d %d", ev->canvas.x - x, ev->canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_UP_DOWN)) {
		LOGD("mouse up disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	if (ev->event_flags & EVAS_EVENT_FLAG_ON_HOLD) {
		LOGD("Event canceld");
		__cancel_touch(toolkit_evas_h);
		return;
	}

	/*
	 * We need to send move event for
	 * consistency with old widget framework
	 */
	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_MOVE,
			0,
			0,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);

	toolkit_evas_h->is_down = false;
	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_UP,
			0,
			ev->button,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_touch_multi_down(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Multi_Down *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("multi down: %d %d", ev->canvas.x - x, ev->canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
				SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_UP_DOWN)) {
		LOGD("multi down disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_touch_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_TOUCH_DOWN,
			ev->device,
			1,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_touch_multi_up(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Multi_Up *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("multi up: %d %d", ev->canvas.x - x, ev->canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
				SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_UP_DOWN)) {
		LOGD("multi up disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_touch_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_TOUCH_UP,
			ev->device,
			1,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_mouse_move(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Move *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;
	int delta_y;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("mouse move: %d %d", ev->cur.canvas.x - x, ev->cur.canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_MOVE_X_Y)) {
		LOGD("mouse move disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	if (toolkit_evas_h->flick) {
		LOGD("flick event");
		return;
	}

	delta_y = ev->cur.canvas.y - y - toolkit_evas_h->prev_y;
	if (delta_y < 0)
		delta_y *= -1;

	if (delta_y > 10) {
		toolkit_evas_h->flick = true;
		return;
	}

	toolkit_evas_h->prev_y = ev->cur.canvas.y - y;
	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_MOVE,
			0,
			0,
			ev->cur.canvas.x - x,
			ev->cur.canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_mouse_wheel(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Wheel *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	struct tizen_remote_surface *surface;

	LOGD("mouse wheel");

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_WHEEL)) {
		LOGD("mouse wheel disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_mouse_wheel(surface,
			ev->direction,
			ev->z,
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_show(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	LOGD("show");
	__obj_update_visibility(data, e, obj, event_info);
}

static void __rs_cb_hide(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	LOGD("hide");
}

static void __rs_cb_resize(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	LOGD("resize");
	__obj_update_visibility(data, e, obj, event_info);
}

static void __toolkit_update_cb(struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
		int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys,
		const char *appid, const char *instance_id, const int pid, void *data)
{
	bool is_added = false;
	Evas_Native_Surface ns;
	tbm_surface_h tbm_surface;
	int width, height;
	char *map;
	char format[] = "png";
	char plug_id[256];
	screen_connector_toolkit_evas_h toolkit_evas_h;
	cur_buffer_info_h c_buf_info;
	bool duplicate_buffer = false;
	Ecore_Wl_Window *wl_win;

	toolkit_evas_h = (screen_connector_toolkit_evas_h)data;

	if (toolkit_evas_h->blocked || toolkit_evas_h->bind_win_id > 0) {
		if (tizen_remote_surface_get_version(trs)
				>= TIZEN_REMOTE_SURFACE_RELEASE_SINCE_VERSION) {
			tizen_remote_surface_release(trs, tbm);
			g_hash_table_remove(__cur_buffer_table, tbm);
		}
		close(img_file_fd); /* close passed fd */
		return;
	}

	if (toolkit_evas_h->img_tbm == NULL)	{
		LOGD("first added !!!! %d, %d %s", (toolkit_evas_h->img_tbm == NULL), type, appid);
		toolkit_evas_h->img_tbm = evas_object_image_filled_add(
				evas_object_evas_get(toolkit_evas_h->type_h->viewer_win));

		/* Image alpha set */
		evas_object_image_alpha_set(toolkit_evas_h->img_tbm, 1);

		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_DOWN,
				__rs_cb_mouse_down, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_UP,
				__rs_cb_mouse_up, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MULTI_DOWN,
				__rs_cb_touch_multi_down, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MULTI_UP,
				__rs_cb_touch_multi_up, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_MOVE,
				__rs_cb_mouse_move, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_WHEEL,
				__rs_cb_mouse_wheel, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_SHOW,
				__rs_cb_show, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_HIDE,
				__rs_cb_hide, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_RESIZE,
				__rs_cb_resize, toolkit_evas_h);

		/* Store toolkit_evas_h */
		g_hash_table_insert(toolkit_evas_h->type_h->toolkit_table, strdup(instance_id), toolkit_evas_h);

		/* Set data to use in accessibility */
		snprintf(plug_id, sizeof(plug_id), "%s:%d", instance_id, pid);
		evas_object_data_set(toolkit_evas_h->img_tbm, "___PLUGID", strdup(plug_id));

		evas_object_data_set(toolkit_evas_h->img_tbm, SC_TOOLKIT_HANDLE_TAG, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOVE, __obj_update_visibility, toolkit_evas_h);

		wl_win = elm_win_wl_window_get(
				toolkit_evas_h->type_h->viewer_win);
		if (!wl_win)
			LOGE("failed to get wl_win");

		__set_window_visibility_info(toolkit_evas_h, ecore_wl_window_id_get(wl_win),
				VISIBILITY_TYPE_UNKNOWN);
	}

	if (toolkit_evas_h->img_type != type)
		is_added = true;
	toolkit_evas_h->img_type = type;

	/* check type of given buffer */
	if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM) {
		/* get tbm surface from buffer */
		tbm_surface = wl_buffer_get_user_data(tbm);
		width = tbm_surface_get_width(tbm_surface);
		height = tbm_surface_get_height(tbm_surface);

		memset(&ns, 0, sizeof(Evas_Native_Surface));
		ns.version = EVAS_NATIVE_SURFACE_VERSION;
		ns.type = EVAS_NATIVE_SURFACE_TBM;
		ns.data.tbm.buffer = tbm_surface;

		evas_object_resize(toolkit_evas_h->img_tbm, width, height);
		evas_object_image_size_set(toolkit_evas_h->img_tbm, width, height);
		evas_object_image_native_surface_set(toolkit_evas_h->img_tbm, &ns); /* set native surface */
		evas_object_image_pixels_dirty_set(toolkit_evas_h->img_tbm, EINA_TRUE); /* set dirty for image updating */
	} else if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
		if (!toolkit_evas_h->img_file)
			toolkit_evas_h->img_file = evas_object_image_filled_add(evas_object_evas_get(toolkit_evas_h->type_h->viewer_win));

		map = mmap(NULL, img_file_size, PROT_READ, MAP_SHARED, img_file_fd, 0);
		evas_object_image_memfile_set(toolkit_evas_h->img_file, map, img_file_size, format, NULL);
		evas_object_image_size_get(toolkit_evas_h->img_tbm, &width, &height);
		evas_object_image_fill_set(toolkit_evas_h->img_file, 0, 0, width, height);
		evas_object_resize(toolkit_evas_h->img_file, width, height);
		munmap(map, img_file_size);

		evas_object_image_native_surface_set(toolkit_evas_h->img_tbm, NULL); /* set null to previous object for the tbm type */
	}

	if (toolkit_evas_h->cur_buffer) {
		if (tizen_remote_surface_get_version(trs) >= TIZEN_REMOTE_SURFACE_RELEASE_SINCE_VERSION) {
			c_buf_info = g_hash_table_lookup(__cur_buffer_table, toolkit_evas_h->cur_buffer);
			if (c_buf_info != NULL) {
				if (c_buf_info->cur_buffer != tbm) {
					__cur_buffer_info_unref(trs, c_buf_info);
				} else {
					duplicate_buffer = true;
					LOGW("Same buffer %p", tbm);
				}
			}
		}
	}

	toolkit_evas_h->cur_buffer = tbm;
	c_buf_info = g_hash_table_lookup(__cur_buffer_table, toolkit_evas_h->cur_buffer);
	if (c_buf_info == NULL) {
		c_buf_info = (cur_buffer_info_h)calloc(1, sizeof(struct _cur_buffer_info_h));
		c_buf_info->cur_buffer = toolkit_evas_h->cur_buffer;
		__cur_buffer_info_ref(c_buf_info);
		g_hash_table_insert(__cur_buffer_table, toolkit_evas_h->cur_buffer, c_buf_info);
	} else {
		if (!duplicate_buffer)
			__cur_buffer_info_ref(c_buf_info);
	}

	if (is_added) {
		LOGD("call added !!!! %s, type %d", appid, type);
		if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
			toolkit_evas_h->ops->added_cb(appid, instance_id, pid, toolkit_evas_h->img_file,
					toolkit_evas_h->data);
			__clear_img_tbm(toolkit_evas_h);
		} else {
			toolkit_evas_h->ops->added_cb(appid, instance_id, pid, toolkit_evas_h->img_tbm,
					toolkit_evas_h->data);
			__clear_img_file(toolkit_evas_h);
		}
	} else {
		if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
			toolkit_evas_h->ops->updated_cb(appid, instance_id, pid, toolkit_evas_h->img_file,
					toolkit_evas_h->data);
			__clear_img_tbm(toolkit_evas_h);
		} else {
			toolkit_evas_h->ops->updated_cb(appid, instance_id, pid, toolkit_evas_h->img_tbm,
					toolkit_evas_h->data);
			__clear_img_file(toolkit_evas_h);
		}
	}

	close(img_file_fd); /* close passed fd */
}

static void __toolkit_added_cb(const char *appid, const char *instance_id, const int pid, void *data)
{
}

static void __toolkit_removed_cb(const char *appid, const char *instance_id, const int pid, void *data)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	char *plug_id;
	GHashTable *toolkit_table;

	LOGD("missing %s %s", appid, instance_id);
	toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	if (!toolkit_evas_h) {
		LOGE("Null toolkit_evas_h");
		return;
	}

	toolkit_table = toolkit_evas_h->type_h->toolkit_table;

	if (toolkit_evas_h->img_tbm != NULL) {
		plug_id = evas_object_data_del(toolkit_evas_h->img_tbm, "___PLUGID");
		if (plug_id)
			free(plug_id);

		evas_object_data_del(toolkit_evas_h->img_tbm, SC_TOOLKIT_HANDLE_TAG);
		toolkit_evas_h->ops->removed_cb(appid, instance_id, pid, toolkit_evas_h->img_tbm, toolkit_evas_h->data);
	} else if (toolkit_evas_h->img_file != NULL) {
		toolkit_evas_h->ops->removed_cb(appid, instance_id, pid, toolkit_evas_h->img_file, toolkit_evas_h->data);
	}

	g_hash_table_remove(toolkit_table, instance_id);
}

EXPORT_API int screen_connector_toolkit_evas_set_delayed_resuming_time(int ms)
{
	__delayed_resuming_time = ms;

	return 0;
}

EXPORT_API screen_connector_toolkit_evas_h screen_connector_toolkit_evas_add_with_win(screen_connector_toolkit_evas_ops *ops, char *id,
		screen_connector_screen_type_e type, Evas_Object *win, void *data)
{
	screen_connector_toolkit_evas_h handle;
	screen_connector_toolkit_evas_ops *evas_ops;
	screen_connector_toolkit_ops toolkit_ops;
	screen_connector_type_evas_h type_h;
	int int_type = type;
	struct wl_surface *surface;
	Ecore_Wl_Window *wl_win;

	if (win == NULL || ops == NULL || id == NULL
		|| type < SCREEN_CONNECTOR_SCREEEN_TYPE_WIDGET
		|| type > SCREEN_CONNECTOR_SCREEEN_TYPE_ALL) {
		LOGE("Invalid param NULL");
		return NULL;
	}

	wl_win = elm_win_wl_window_get(win);
	if (wl_win == NULL) {
		LOGE("failed to get wl_win");
		return NULL;
	}

	surface = ecore_wl_window_surface_get(wl_win);
	if (surface == NULL) {
		LOGE("failed to get surface");
		return NULL;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(int_type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	toolkit_ops.added_cb = __toolkit_added_cb;
	toolkit_ops.updated_cb = NULL;
	toolkit_ops.removed_cb = __toolkit_removed_cb;

	evas_ops = (screen_connector_toolkit_evas_ops *)calloc(1, sizeof(screen_connector_toolkit_evas_ops));
	if (evas_ops == NULL) {
		LOGE("evas_ops calloc fail");
		return NULL;
	}

	memcpy(evas_ops, ops, sizeof(screen_connector_toolkit_evas_ops));
	handle = (screen_connector_toolkit_evas_h)calloc(1, sizeof(struct _screen_connector_toolkit_evas_h));
	if (handle == NULL) {
		LOGE("handle calloc fail");
		free(evas_ops);
		return NULL;
	}
	handle->img_type = -1;
	handle->ops = evas_ops;
	handle->data = data;
	handle->type_h = type_h;
	handle->toolkit_h = screen_connector_toolkit_add_with_surface(&toolkit_ops, id, type, surface, handle);
	handle->screen_type = type;

	return handle;
}

EXPORT_API screen_connector_toolkit_evas_h screen_connector_toolkit_evas_add(screen_connector_toolkit_evas_ops *ops, char *id,
		screen_connector_screen_type_e type, void *data)
{
	screen_connector_toolkit_evas_h handle;
	screen_connector_toolkit_evas_ops *evas_ops;
	screen_connector_toolkit_ops toolkit_ops;
	screen_connector_type_evas_h type_h;
	int int_type = type;

	if (screen_connector_toolkit_is_exist(id, type)) {
		LOGE("Already exist ID : %s", id);
		return NULL;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(int_type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	toolkit_ops.added_cb = __toolkit_added_cb;
	toolkit_ops.updated_cb = __toolkit_update_cb;
	toolkit_ops.removed_cb = __toolkit_removed_cb;

	evas_ops = (screen_connector_toolkit_evas_ops *)calloc(1, sizeof(screen_connector_toolkit_evas_ops));
	if (evas_ops == NULL) {
		LOGE("evas_ops calloc fail");
		return NULL;
	}

	memcpy(evas_ops, ops, sizeof(screen_connector_toolkit_evas_ops));
	handle = (screen_connector_toolkit_evas_h)calloc(1, sizeof(struct _screen_connector_toolkit_evas_h));
	if (handle == NULL) {
		LOGE("handle calloc fail");
		free(evas_ops);
		return NULL;
	}
	handle->img_type = -1;
	handle->ops = evas_ops;
	handle->data = data;
	handle->type_h = type_h;
	handle->toolkit_h = screen_connector_toolkit_add(&toolkit_ops, id, type, handle);
	handle->screen_type = type;

	return handle;
}

EXPORT_API int screen_connector_toolkit_evas_remove(screen_connector_toolkit_evas_h handle)
{
	const char *instance_id;

	if (handle == NULL)
		return -1;

	instance_id = screen_connector_toolkit_get_instance_id(handle->toolkit_h);
	if (instance_id == NULL) {
		LOGE("Null instance_id");
		return -1;
	}

	g_hash_table_remove(handle->type_h->toolkit_table, instance_id);

	__destroy_toolkit_evas_h(handle);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_update(const char *appid, const char *instance_id)
{
	return screen_connector_toolkit_update(appid, instance_id);
}

static void __send_visibility(gpointer key, gpointer value, gpointer user_data)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)value;
	Ecore_Wl_Event_Window_Visibility_Change *ev = user_data;
	int ret;
	visibility_type type;
	GList *iter;
	viewer_visibility_h visibility_h;

	if (toolkit_evas_h->freeze)
		return;

	if (!__obj_is_visible(toolkit_evas_h))
		return;

	if (ev->fully_obscured)
		type = VISIBILITY_TYPE_FULLY_OBSCURED;
	else
		type = VISIBILITY_TYPE_UNOBSCURED;


	__set_window_visibility_info(toolkit_evas_h, ev->win, type);
	type = VISIBILITY_TYPE_FULLY_OBSCURED;
	iter = toolkit_evas_h->viewer_visibility_list;
	while (iter != NULL) {
		visibility_h = (viewer_visibility_h)iter->data;
		if (visibility_h->visibility == VISIBILITY_TYPE_UNKNOWN) {
			LOGE("viewer not ready");
			return;
		} else if (visibility_h->visibility ==
				VISIBILITY_TYPE_UNOBSCURED) {
			type = VISIBILITY_TYPE_UNOBSCURED;
			break;
		}
		iter = g_list_next(iter);
	}

	LOGW("set visibility to %d", type);
	ret = __set_visibility(toolkit_evas_h, type);
	toolkit_evas_h->auto_visibility = type;
	if (ret == -1)
		LOGE("failed to set object visibility  set %p to %d", toolkit_evas_h->img_tbm, type);
}

static Eina_Bool __visibility_cb(void *data, int type, void *event)
{
	Ecore_Wl_Event_Window_Visibility_Change *ev = event;
	GHashTableIter iter;
	gpointer key, value;
	screen_connector_type_evas_h cur_type;

	LOGD("visibility change: %d %d", (unsigned int)ev->win,
			(unsigned int)ev->fully_obscured);

	if (!__type_table)
		return ECORE_CALLBACK_RENEW;

	g_hash_table_iter_init(&iter, __type_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		cur_type = (screen_connector_type_evas_h)value;
		if (cur_type->toolkit_table)
			g_hash_table_foreach(cur_type->toolkit_table,
					__send_visibility, ev);
	}

	return ECORE_CALLBACK_RENEW;
}

EXPORT_API int screen_connector_toolkit_evas_start_visibility_notify(void)
{
	if (__visibility_listener)
		return 0;

	__visibility_listener = ecore_event_handler_add(ECORE_WL_EVENT_WINDOW_VISIBILITY_CHANGE, __visibility_cb, NULL);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_stop_visibility_notify(void)
{
	if (!__visibility_listener)
		return 0;

	ecore_event_handler_del(__visibility_listener);
	__visibility_listener = NULL;

	return 0;
}

screen_connector_toolkit_evas_h __find_toolkit_evas_h(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = evas_object_data_get(obj, SC_TOOLKIT_HANDLE_TAG);

	return toolkit_evas_h;
}

EXPORT_API int screen_connector_toolkit_evas_freeze_visibility(Evas_Object *obj, visibility_type type)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}
	toolkit_evas_h->freeze = 1;

	return __set_visibility(toolkit_evas_h, type);
}

EXPORT_API int screen_connector_toolkit_evas_thaw_visibility(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}
	toolkit_evas_h->freeze = 0;

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_send_visibility(Evas_Object *obj, visibility_type type)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}

	return __set_visibility(toolkit_evas_h, type);
}

EXPORT_API int screen_connector_toolkit_evas_send_mouse_up(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	int x;
	int y;
	int w;
	int h;
	unsigned int timestamp;
	struct tizen_remote_surface *surface;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object. not widget object.");
		return -1;
	}

	if (!toolkit_evas_h->img_tbm) {
		LOGE("surface not yet initialized.");
		return -1;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return -1;
	}

	evas_object_geometry_get(obj, &x, &y, &w, &h);
	timestamp = (unsigned int)((unsigned long long)(ecore_time_get() * 1000.0) & 0xffffffff);
	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_UP,
			0,
			1,
			w / 2,
			h / 2,
			0,
			0,
			0,
			0,
			TIZEN_INPUT_DEVICE_CLAS_MOUSE,
			TIZEN_INPUT_DEVICE_SUBCLAS_NONE,
			"",
			timestamp);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_send_touch_cancel(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object. not widget object.");
		return -1;
	}

	if (toolkit_evas_h->img_tbm) {
		__cancel_touch(toolkit_evas_h);
	} else {
		LOGE("Null tbm image.");
		return -1;
	}

	return 0;
}

EXPORT_API screen_connector_toolkit_evas_h screen_connector_toolkit_evas_add_by_rid(screen_connector_toolkit_evas_ops *ops,
	int res_id, screen_connector_screen_type_e type, void *data)
{
	screen_connector_toolkit_evas_h handle;
	screen_connector_toolkit_evas_ops *evas_ops;
	screen_connector_toolkit_ops toolkit_ops;
	screen_connector_type_evas_h type_h;
	char id_str[70];
	char uuid[37];
	uuid_t u;

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	toolkit_ops.added_cb = __toolkit_added_cb;
	toolkit_ops.updated_cb = __toolkit_update_cb;
	toolkit_ops.removed_cb = __toolkit_removed_cb;

	evas_ops = (screen_connector_toolkit_evas_ops *)calloc(1, sizeof(screen_connector_toolkit_evas_ops));
	if (evas_ops == NULL) {
		LOGE("evas_ops calloc fail");
		return NULL;
	}

	memcpy(evas_ops, ops, sizeof(screen_connector_toolkit_evas_ops));
	handle = (screen_connector_toolkit_evas_h)calloc(1, sizeof(struct _screen_connector_toolkit_evas_h));
	if (handle == NULL) {
		LOGE("handle calloc fail");
		free(evas_ops);
		return NULL;
	}
	handle->img_type = -1;
	handle->ops = evas_ops;
	handle->data = data;
	handle->type_h = type_h;
	handle->screen_type = type;

	uuid_generate(u);
	uuid_unparse(u, uuid);
	snprintf(id_str, sizeof(id_str), "%s:%d", uuid, res_id);
	handle->toolkit_h = screen_connector_toolkit_add_by_rid(&toolkit_ops, id_str, type, res_id, handle);

	return handle;
}

EXPORT_API int screen_connector_toolkit_evas_get_rid(Evas_Object *obj, int *resource_id)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object.");
		return -1;
	}
	*resource_id = screen_connector_toolkit_get_surface_id(toolkit_evas_h->toolkit_h);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_get_pid(Evas_Object *obj, int *pid)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	int toolkit_pid = 0;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object.");
		return -1;
	}

	toolkit_pid = screen_connector_toolkit_get_pid(toolkit_evas_h->toolkit_h);
	if (toolkit_pid == 0) {
		LOGE("Invalid object");
		return -1;
	}

	*pid = toolkit_pid;

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_is_visible(Evas_Object *obj, bool *is_visible)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	if (obj == NULL || is_visible == NULL) {
		LOGE("invalid argument");
		return -1;
	}
	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}

	*is_visible = __obj_is_visible(toolkit_evas_h);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_bind(screen_connector_toolkit_evas_h h,
		Evas_Object *win)
{
	struct wl_surface *surface;
	Ecore_Wl_Window *wl_win;
	struct tizen_remote_surface *trs;
	cur_buffer_info_h c_buf_info;
	int ret;
	int bind_win_id;

	if (!h)
		return -1;

	wl_win = elm_win_wl_window_get(win);
	if (!wl_win) {
		LOGE("failed to get wl_win");
		return -1;
	}

	surface = ecore_wl_window_surface_get(wl_win);
	if (!surface) {
		LOGE("failed to get surface");
		return -1;
	}

	bind_win_id = ecore_wl_window_id_get(wl_win);
	ret = __set_window_visibility_info(h, bind_win_id,
			VISIBILITY_TYPE_UNKNOWN);
	if (ret < 0) {
		LOGE("failed to set visibility info");
		return -1;
	}
	h->bind_win_id = bind_win_id;

	__cancel_touch(h);
	if (h->cur_buffer) {
		trs = screen_connector_toolkit_get_trs(h->toolkit_h);
		if (trs == NULL) {
			LOGE("failed to get tizen remote surface");
			return -1;
		}

		if (tizen_remote_surface_get_version(trs)
				>= TIZEN_REMOTE_SURFACE_RELEASE_SINCE_VERSION) {
			c_buf_info = g_hash_table_lookup(__cur_buffer_table,
					h->cur_buffer);
			if (c_buf_info != NULL)
				__cur_buffer_info_unref(trs, c_buf_info);
		}
	}

	return screen_connector_toolkit_bind(h->toolkit_h, surface);
}

EXPORT_API int screen_connector_toolkit_evas_unbind(screen_connector_toolkit_evas_h h)
{
	if (!h || !(h->toolkit_h)) {
		LOGE("invalid argument %p", h);
		return -1;
	}

	__remove_window_visibility_info(h, h->bind_win_id);
	h->bind_win_id = 0;
	return screen_connector_toolkit_unbind(h->toolkit_h);
}

EXPORT_API visibility_type screen_connector_toolkit_evas_auto_visibility(
		screen_connector_toolkit_evas_h h)
{
	if (!h) {
		LOGE("invalid argument %p", h);
		return -1;
	}

	return h->auto_visibility;
}

EXPORT_API int screen_connector_toolkit_evas_set_changed_event_filter(
		screen_connector_toolkit_evas_h toolkit_evas_h,
		screen_connector_evas_changed_event_filter_type type)
{
	cur_buffer_info_h c_buf_info;
	struct tizen_remote_surface *trs;
	int ret;

	ret = screen_connector_toolkit_set_changed_event_filter(
			toolkit_evas_h->toolkit_h, type);

	if (ret < 0) {
		LOGE("failed to set changed event filter");
		return ret;
	}

	if (type != SCREEN_CONNECTOR_EVAS_CHANGED_EVENT_FILTER_ALL &&
			type != SCREEN_CONNECTOR_EVAS_CHANGED_EVENT_FILTER_TBM) {
		LOGD("filter = %d", type);
		toolkit_evas_h->blocked = false;
		return ret;
	}

	__cancel_touch(toolkit_evas_h);
	toolkit_evas_h->blocked = true;
	if (toolkit_evas_h->cur_buffer) {
		trs = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
		if (trs == NULL) {
			LOGE("failed to get tizen remote surface");
			return -1;
		}

		if (tizen_remote_surface_get_version(trs)
				>= TIZEN_REMOTE_SURFACE_RELEASE_SINCE_VERSION) {
			c_buf_info = g_hash_table_lookup(__cur_buffer_table,
					toolkit_evas_h->cur_buffer);
			if (c_buf_info != NULL)
				__cur_buffer_info_unref(trs, c_buf_info);
		}
	}

	return 0;
}
